import paho.mqtt.client as PahoMQTT
import json
import subprocess
import random
import logging
from logging.handlers import RotatingFileHandler
import threading
import time
import ast
import can
from can.interfaces import socketcan
import struct

confFilename = "MQTT-conf_realtime.json"


def CANFactory(RTStart, CANStart, canIf, bus):
	'''This threading function initializes the CAN bus and wait for the event.
	When it starts it takes the data filtered and it proceed to create the realtime streaming.
	
	'''

	global buffer
	#global status
	global can_filters
	


	while True:
		RTStart.wait()

		if CANStart.is_set():
	
			bus.set_filters(can_filters["can_filters"])		
			logging.debug("Start CAN listeners with filters:\n{}".format(can_filters))

			ext_filters = can_filters["ext_filters"]
			
			#status = []
			buffer = []

			for var in ext_filters["variables"]:
				buffer.append(None)
				#status.append(False)

		msg = bus.recv()
		logging.debug("Message IGNORED for CANStart False")

		while RTStart.is_set() and CANStart.is_set():
			#receive msg
			msg = bus.recv()
			logging.debug("Msg received:\t{}\t{}".format(hex(msg.arbitration_id), msg.data))

			'''
			for pgn in filters_dict.keys():
				if pgn in format(msg.arbitration_id, '0>16X'):
			'''

			unpack_data = format(struct.unpack('<Q', msg.data)[0], '0>16X')
			#logging.debug("unpack data        <{}>".format(unpack_data))

			pgn = msg.arbitration_id #format(msg.arbitration_id, '0>8X')
			#logging.debug("msg arbitration id <{}>".format(pgn))

			

			for var in ext_filters["variables"]:
				if pgn == var["can_id"]:
					#logging.debug("pgn matched with <{}>".format(var["can_id"]))
					data = int(format(int(unpack_data, base=16), '0>64b')[-var["toBit"]-1:None if not var["fromBit"] else -var["fromBit"]], base=2)
					#logging.debug("data extracted <{}>".format(data))
					buffer[ext_filters["variables"].index(var)] = data
					#logging.debug("buffer updated <{}>".format(buffer))
					#status[ext_filters["variables"].index(var)] = True






			#data = unpack_data[filters_dict[pgn]["byte"][0]<<1:filters_dict[pgn]["byte"][1]<<1]
			#logging.debug("data byte  <{}>".format(data))

			#size = len(data)<<2

			#data = format(int(data, base=16), '0>{}b'.format(size))[filters_dict[pgn]["fromBit"]:filters_dict[pgn]["toBit"]]
			#logging.debug("data bin:\tsize <{}>\tdata <{}>".format(size,data))

			'''
			with bufferLock:
				logging.debug("STATUSLOCK -> ACQUIRED")
          
				buffer[pgn] = data
				logging.debug("buffer updated:   {}".format(buffer))
          
				status[filters_dict[pgn]["start"]] = True
				logging.debug("status:  {}".format(status))
				logging.info("STATUSLOCK -> RELEASED")
			'''
	 	

			'''




			val1 = random.random() * 100
			val2 = random.randrange(800, 6000)
			val3 = random.random() * 100

			jsonout = {"temperature": val1,
								"engine_speed": val2,
								"gravel_speed": val3}

			logging.debug("Data ready to send: {}".format(jsonout))

			pub.Publish(str(jsonout))

			time.sleep(1)
			'''
		#time.sleep(1)


def periodicPublisher(RTStart, CANStart, publisher, interval=2.0):
	'''Thread function that periodically extract the raw data from the can buffer,
	it tranfroms them into human-readable data and publish with the correct topic'''

	global buffer
	global can_filters

	time.sleep(2)
	
	while True:
		RTStart.wait()
		payload_out = {}
		
		while RTStart.is_set() and CANStart.is_set():
			for var in can_filters["ext_filters"]["variables"]:
				payload_out[var["name"]] = None
			
			#logging.debug("payload_out after init:\n{}".format(payload_out))
			logging.debug("len(buffer) <{}>".format(len(buffer)))
			logging.debug("buffer <{}>".format(buffer))

			for i in range(len(buffer)):
				'''
				try: 
					data = int(buffer[i], base=2)
				except TypeError:
					payload_out[can_filters["ext_filters"]["variables"][i]["name"]] = None 
					logging.debug("Payload_out update <{}>".format(payload_out))

				else:
				'''
				if buffer[i] != None:
					data = buffer[i]
					logging.debug("i-cycle <{}>\t data int <{}>".format(i,data))
					data *= can_filters["ext_filters"]["variables"][i]["factor"]
					logging.debug("data int *= <{}>".format(data))
					data += can_filters["ext_filters"]["variables"][i]["offset"]
					logging.debug("data int += <{}>".format(data))
					payload_out[can_filters["ext_filters"]["variables"][i]["name"]] = data 
					logging.debug("Payload_out update <{}>".format(payload_out))

			publisher.Publish(str(payload_out))

			time.sleep(interval)
		time.sleep(1)

def keyOnManager (CANStart):
	'''Thread function that check in background the key status and suspends the publishing
	if enabled while the status remains off
	'''

	while True:
		req_getKey = subprocess.Popen("udpctrl in-get 0", stderr=subprocess.PIPE, shell=True)
		getKey = 1 if req_getKey.stderr.read().decode('utf-8').split("=")[1][:-1] == ' on' else 0

		#logging.debug("Key status <{}>".format(getKey))

		if not getKey:
			if CANStart.is_set():
				#logging.debug("Key status <{}>".format(getKey))

				CANStart.clear()
				logging.warning("Stop CAN acquisition due to key switch OFF")
		else:
			if not CANStart.is_set():
				#logging.debug("Key status <{}>".format(getKey))

				CANStart.set()
				logging.warning("Start CAN acquisition due to key switch ON")

		time.sleep(0.5)






def config_extract():
	configurations = {}
	"""Open and extract all the MQTT config"""
	try:
		with open(confFilename, 'r') as f:
			configurations = json.load(f)
	except FileNotFoundError as e:
		raise(e)

		return configurations

 
class MQTTService:
		def __init__(self, configFile, name, eventStartRT, logging):
			f_conf=open(configFile)
			init_conf=json.load(f_conf)
			f_conf.close()
			self._paho_mqtt=PahoMQTT.Client(client_id=name,clean_session=False)
			self._paho_mqtt.on_connect=self.OnConnect
			self._paho_mqtt.on_message=self.OnMessageReceived
			self._paho_mqtt.on_publish=self.OnPublish
			self._paho_mqtt.on_subscribe=self.OnSubscribe


			self.logging=logging
			#self.canBus=bus 

			self.broker=init_conf["mqtt"]["brokerID"]
			self.port=init_conf["mqtt"]["brokerPort"]
			self.keepalive=init_conf["mqtt"]["keepalive"]

			self.LTW_topic= name + init_conf["mqtt"]["LTW_topic"]
			self.sub_topic= name + init_conf["mqtt"]["sub_topic"]
			self.pub_topic= name + init_conf["mqtt"]["pub_topic"]

			self.call_set = name + init_conf["mqtt"]["call_set"]
			self.call_upgrade = name + init_conf["mqtt"]["call_upgrade"]
			self.call_offline = name + init_conf["mqtt"]["call_offline"]


			self._paho_mqtt.message_callback_add(sub=self.call_set, callback=self.OnSetStream)
			self._paho_mqtt.message_callback_add(sub=self.call_offline, callback=self.OnStopStream)

			self.name=name
			self.RTStart=eventStartRT

			self.unconnect_flag=True
			
		def start (self):
			# testament and will
			exception = True
			self._paho_mqtt.will_set(topic=self.LTW_topic,qos=2)

			#set reconnection delay between retry
			self._paho_mqtt.reconnect_delay_set(min_delay=1, max_delay=32)

			#manage connection to broker
			while self.unconnect_flag:
				if exception:
					try:
						self._paho_mqtt.connect(host=self.broker, port=self.port, keepalive=self.keepalive)
				
					except:
						pass
					else:
						exception = False

				time.sleep(2)

			self._paho_mqtt.loop_start()

			# subscribe for a topic
			self._paho_mqtt.subscribe(topic=self.sub_topic)
			

		def stop (self):
			self._paho_mqtt.unsubscribe(topic)
			self._paho_mqtt.loop_stop()
			self._paho_mqtt.disconnect()


		def OnConnect (self, paho_mqtt, userdata, flags, rc):
			self.logging.info("%s connected to %s with result code: %d" % (self.name,self.broker, rc))
			#print ("Subscriber connected to %s with result code: %d" % (self.broker, rc))

			if rc == 0:
				self.unconnect_flag = False
			elif rc == 1:
				self.logging.warning("Connection refused due to incorrect protocol version")
			elif rc == 2:
				self.logging.warning("Connection refused due to invalid client identifier")
			elif rc == 3:
				self.logging.warning("Connection refused due to server unavailable")
			elif rc == 4:
				self.logging.warning("Connection refused due to bad username or password")
			elif rc == 5:
				self.logging.warning("Connection refused due to not authorised")
			else:
				self.logging.warning("connection refused due to unknown reason")



		def OnMessageReceived (self, paho_mqtt , userdata, msg):
			# A new message is received
			self.logging.info("Unknown message received")
			self.logging.debug("Topic:<" + msg.topic+">, QoS: <"+str(msg.qos)+"> Message: <"+str(msg.payload) + ">")
			#print ("Topic:'" + msg.topic+"', QoS: '"+str(msg.qos)+"' Message: '"+str(msg.payload) + "'")
			#global output
			self.Publish(data=msg.topic, topic=self.pub_topic + "message_unknown")

		def OnSetStream (self, paho_mqtt , userdata, msg):
			'''Callback function when a new realtime stream setting is arrived.
			It privides the extraction of the dictionary and the set of the Lock event
			for other thread function waiting to start
			'''

			self.logging.info("Setting new CAN filters")
			self.logging.debug("Topic:<" + msg.topic+">, QoS: <"+str(msg.qos)+"> Message: <"+str(msg.payload) + ">")

			global can_filters
			
			#extract filters
			can_filters=ast.literal_eval(msg.payload.decode())
			

			#set filters
			#self.canBus.set_filters(filters=ext_filters)

			#start CAN factory
			self.RTStart.set()

		def OnStopStream (self):
			'''Callback function when the client listening for the realtime stream 
			disconnects ungracefully and so the stream has to be stop
			'''

			self.logging.info("Client disconnected ungracefully: stop realtime")
			self.RTStart.clear()


		def OnUpdate (self):
			'''Callback function when the application needs to be updated. It launches the automatic update scripting
			that download the application and restart the service
			'''
			_ = subprocess.Popen("sh MQTTService_update.sh", shell=True)



		def Publish(self, data, topic=self.sub_topic):		
			self._paho_mqtt.publish(topic, data)   #### data=> dictionary "value": int|float|string

		def OnPublish(self, paho_mqtt, userdata, mid):
			#A new data is published
			logging.info("Published new data")
			#logging.debug("Topic: {}".format(self.pub_topic))

		def OnSubscribe(self, paho_mqtt, userdata, mid, qos):
			#A new subscription is made
			self.logging.info("Subscribed topic:  {}".format(self.sub_topic))
					


'''TODO: manage this error when call connect
sock = self._create_socket_connection()
  File "/usr/local/lib/python3.5/dist-packages/paho/mqtt/client.py", line 3685, in _create_socket_connection
    return socket.create_connection(addr, timeout=self._connect_timeout, source_address=source)
  File "/usr/lib/python3.5/socket.py", line 712, in create_connection
    raise err
  File "/usr/lib/python3.5/socket.py", line 703, in create_connection
    sock.connect(sa)
socket.timeout: timed out
'''
					
if __name__ == '__main__':

	f_conf=open("MQTT-conf_realtime.json")
	init_conf=json.load(f_conf)
	f_conf.close()

	handler = RotatingFileHandler("serviceMQTT.log", mode='a', maxBytes=500000, backupCount=1)
	logging.basicConfig(format='%(asctime)s - %(funcName)s - %(message)s', level="DEBUG", handlers=[handler])
	#logging.basicConfig(filename='/var/log/mqtt.log', filemode='a', format='%(asctime)s - %(levelname)s - %(funcName)s - %(message)s', level='DEBUG')	#tail -f syslog.log


	#take the IMEI sn from sybok
	req_ID = subprocess.Popen("udpctrl -alt property-read F112", stderr=subprocess.PIPE, shell=True)
	ID = req_ID.stderr.read().decode('utf-8').split("=")[1][:-1]

	#init the can if
	canIf = init_conf["can"]["channel"]

	buffer = []
	payload_out = {}

	#threading event to start/stop the CAN processing data
	eventStartRT = threading.Event()
	eventStartCAN = threading.Event()

	with can.interfaces.socketcan.SocketcanBus(channel=canIf, receive_own_messages=False, fd=False) as bus:
		logging.info("SocketCAN open on channel: {}".format(canIf))

		ser = MQTTService("MQTT-conf_realtime.json",ID,eventStartRT, logging)		
		ser.start()									

		#Thread initialization
		threadCAN = threading.Thread(target=CANFactory, args=(eventStartRT,eventStartCAN, canIf, bus,))
		threadPublisher = threading.Thread(target=periodicPublisher, args=(eventStartRT,eventStartCAN,ser,init_conf["mqtt"]["periodic_send"],))
		threadKeyManager = threading.Thread(target=keyOnManager, args=(eventStartCAN,))
		threadCAN.start()
		threadPublisher.start()
		threadKeyManager.start()
		#blocking function that wait undefinetely
		threadCAN.join()
		threadPublisher.join()
		threadKeyManager.join()





		



